#coding: utf-8

# Copyright (c) 2015-2016, thumbor-community
# Use of this source code is governed by the MIT license that can be
# found in the LICENSE file.
from os.path import splitext

from json import dumps, loads

from botocore.exceptions import BotoCoreError, ClientError

from thumbor.storages import BaseStorage
from thumbor.utils import logger

from ..aws.storage import AwsStorage

class Storage(AwsStorage, BaseStorage):
    """
    S3 Storage
    """
    def __init__(self, context):
        """
        Constructor
        :param Context context: Thumbor's context
        """
        BaseStorage.__init__(self, context)
        AwsStorage.__init__(self, context, 'TC_AWS_STORAGE')
        self.storage_expiration_seconds = context.config.get('STORAGE_EXPIRATION_SECONDS', 3600)


    async def put(self, path, file_bytes):
        """
        Stores image
        :param string path: Path to store data at
        :param bytes bytes: Data to store
        :rtype: string
        """
        try:
            await self._put_object(file_bytes, self._normalize_path(path))
        except BotoCoreError as err:
            logger.exception('Unable to store object: %s', err)
            return None

        return path

    async def put_crypto(self, path):
        """
        Stores crypto data at given path
        :param string path: Path to store the data at
        :return: Path where the crypto data is stored
        """
        if not self.context.config.STORES_CRYPTO_KEY_FOR_EACH_IMAGE:
            return

        if not self.context.server.security_key:
            raise RuntimeError(
                "STORES_CRYPTO_KEY_FOR_EACH_IMAGE can't be "
                "True if no SECURITY_KEY specified"
            )

        file_abspath = self._normalize_path(path)
        crypto_path = '%s.txt' % splitext(file_abspath)[0]

        try:
            await self._put_object(self.context.server.security_key.encode('utf-8'), crypto_path)
        except BotoCoreError as err:
            logger.exception('Unable to store crypto object: %s', err)
            return None

        logger.debug(
            "Stored crypto at %s (security key: %s)",
            crypto_path,
            self.context.server.security_key,
        )

        return file_abspath

    async def put_detector_data(self, path, data):
        """
        Stores detector data at given path
        :param string path: Path to store the data at
        :param string data:  Data to store
        :return: Path where the data is stored
        :rtype: string
        """
        file_abspath = self._normalize_path(path)

        path = '%s.detectors.txt' % splitext(file_abspath)[0]

        try:
            await self._put_object(dumps(data).encode('utf-8'), path)
        except BotoCoreError as err:
            logger.exception('Unable to store detector data: %s', err)
            return None

        return file_abspath

    async def get_crypto(self, path):
        """
        Retrieves crypto data at path
        :param string path: Path to search for crypto data
        """
        file_abspath = self._normalize_path(path)
        crypto_path = "%s.txt" % (splitext(file_abspath)[0])

        try:
            file_key = await self.storage.get(crypto_path)
        except ClientError as err:
            logger.warn("[STORAGE] s3 key not found at %s" % crypto_path)
            return None

        async with file_key['Body'] as stream:
            file_key = await stream.read()

        return file_key.decode('utf-8')

    async def get_detector_data(self, path):
        """
        Retrieves detector data from storage
        :param string path: Path where the data is stored
        """
        file_abspath = self._normalize_path(path)
        path = '%s.detectors.txt' % splitext(file_abspath)[0]

        try:
            file_key = await self.storage.get(path)
        except ClientError:
            return None

        if not file_key or self.is_expired(file_key) or 'Body' not in file_key:
            return None

        async with file_key['Body'] as stream:
            return loads(await stream.read())

    async def get(self, path):
        """
        Gets data at path
        :param string path: Path for data
        """

        try:
            file = await super(Storage, self).get(path)
        except BotoCoreError:
            return None

        async with file['Body'] as stream:
            return await stream.read()

    async def exists(self, path):
        """
        Tells if data exists at given path
        :param string path: Path to check
        """
        file_abspath = self._normalize_path(path)
        return await self.storage.exists(file_abspath)

    async def remove(self, path):
        """
        Deletes data at path
        :param string path: Path to delete
        """
        return await self.storage.delete(self._normalize_path(path))




# coding: utf-8

# Copyright (c) 2015, thumbor-community
# Use of this source code is governed by the MIT license that can be
# found in the LICENSE file.

import logging
import os
import signal
import subprocess as sp
import sys
import time

import botocore.session
import mock
import requests
from aiobotocore.session import AioSession
from tornado.testing import AsyncTestCase

from tc_aws.aws.bucket import Bucket
from tests.fixtures.storage_fixture import s3_bucket

logging.basicConfig(level=logging.CRITICAL)

os.environ["TEST_SERVER_MODE"] = "true"

_proxy_bypass = {
    "http": None,
    "https": None,
}


def start_service(host, port):
    args = [sys.executable, "-m", "moto.server", "-H", host,
            "-p", str(port)]

    process = sp.Popen(args, stderr=sp.PIPE)
    url = "http://{host}:{port}".format(host=host, port=port)

    for i in range(0, 30):
        if process.poll() is not None:
            process.communicate()
            break

        try:
            # we need to bypass the proxies due to monkeypatches
            requests.get(url, timeout=0.5)
            break
        except requests.exceptions.ConnectionError:
            time.sleep(0.5)
    else:
        stop_process(process)

    return process


def stop_process(process):
    try:
        process.send_signal(signal.SIGTERM)
        process.communicate()
    except Exception:
        process.kill()
        outs, errors = process.communicate()
        exit_code = process.returncode
        msg = "Child process finished {} not in clean way: {} {}" \
            .format(exit_code, outs, errors)
        raise RuntimeError(msg)


class FakeSession(AioSession):

    def create_client(self, *args, **kwargs):
        if kwargs['endpoint_url'] is None:
            kwargs['endpoint_url'] = "http://localhost:5000"
        return super(FakeSession, self).create_client(*args, **kwargs)


class S3MockedAsyncTestCase(AsyncTestCase):
    _process = None

    @classmethod
    def setUpClass(cls):
        super(S3MockedAsyncTestCase, cls).setUpClass()
        cls._process = start_service("localhost", 5000)

        os.environ['AWS_SHARED_CREDENTIALS_FILE'] = ''
        os.environ['AWS_ACCESS_KEY_ID'] = 'test-key'
        os.environ['AWS_SECRET_ACCESS_KEY'] = 'test-secret-key'
        os.environ['AWS_SESSION_TOKEN'] = 'test-session-token'

    @classmethod
    def tearDownClass(cls):
        super(S3MockedAsyncTestCase, cls).tearDownClass()
        stop_process(cls._process)

    def setUp(self):
        super(S3MockedAsyncTestCase, self).setUp()

        requests.post("http://localhost:5000/moto-api/reset")

        client = botocore.session.get_session().create_client('s3', endpoint_url='http://localhost:5000/')
        client.create_bucket(Bucket=s3_bucket)

        client_patcher = mock.patch('aiobotocore.session.AioSession', FakeSession)
        client_patcher.start()

        self.addCleanup(client_patcher.stop)

    def tearDown(self):
        super(S3MockedAsyncTestCase, self).tearDown()
        # singleton Bucket holds old IOLoop instance which closed after each test
        # this cleans singleton
        Bucket._instances = {}



# coding: utf-8

# Copyright (c) 2015, thumbor-community
# Use of this source code is governed by the MIT license that can be
# found in the LICENSE file.

from os.path import join, abspath, dirname

from thumbor.context import ServerParameters

s3_bucket = 'thumbor-images-test'

IMAGE_URL = 's.glbimg.com/some/image_%s.jpg'
IMAGE_PATH = join(abspath(dirname(__file__)), 'image.jpg')

with open(IMAGE_PATH, 'rb') as img:
    IMAGE_BYTES = img.read()


def get_server(key=None):
    server_params = ServerParameters(8888, 'localhost', 'thumbor.conf', None, 'info', None)
    server_params.security_key = key
    return server_params



# se!/usr/bin/python
# -*- coding: utf-8 -*-
# coding: utf-8

# Copyright (c) 2015, thumbor-community
# Use of this source code is governed by the MIT license that can be
# found in the LICENSE file.

from datetime import datetime, timedelta
from unittest import TestCase
from pytest import raises

from dateutil.tz import tzutc
from thumbor.config import Config
from thumbor.context import Context, RequestParameters
from tornado.testing import gen_test

from .fixtures.storage_fixture import IMAGE_URL, IMAGE_BYTES, get_server, s3_bucket
from tc_aws.storages.s3_storage import Storage
from tests import S3MockedAsyncTestCase


class S3StorageTestCase(S3MockedAsyncTestCase):

    @gen_test
    async def test_can_store_image(self):
        config = Config(TC_AWS_STORAGE_BUCKET=s3_bucket)
        storage = Storage(Context(config=config, server=get_server('ACME-SEC')))

        await storage.put(IMAGE_URL % '1', IMAGE_BYTES)
        topic = await storage.get(IMAGE_URL % '1')

        self.assertEqual(topic, IMAGE_BYTES)

    @gen_test
    async def test_can_get_image_existance(self):
        config = Config(TC_AWS_STORAGE_BUCKET=s3_bucket)
        storage = Storage(Context(config=config, server=get_server('ACME-SEC')))

        await storage.put(IMAGE_URL % '3', IMAGE_BYTES)
        topic = await storage.exists(IMAGE_URL % '3')

        self.assertTrue(topic)

    @gen_test
    async def test_can_get_image_inexistance(self):
        config = Config(TC_AWS_STORAGE_BUCKET=s3_bucket)
        storage = Storage(Context(config=config, server=get_server('ACME-SEC')))

        topic = await storage.exists(IMAGE_URL % '9999')

        self.assertFalse(topic)

    @gen_test
    async def test_can_remove_instance(self):
        config = Config(TC_AWS_STORAGE_BUCKET=s3_bucket,TC_AWS_STORAGE_ROOT_PATH='nana')
        storage = Storage(Context(config=config, server=get_server('ACME-SEC')))
        await storage.put(IMAGE_URL % '4', IMAGE_BYTES)
        await storage.remove(IMAGE_URL % '4')
        topic = await storage.exists(IMAGE_URL % '4')

        self.assertFalse(topic)

    @gen_test
    async def test_can_remove_then_put_image(self):
        config = Config(TC_AWS_STORAGE_BUCKET=s3_bucket)
        storage = Storage(Context(config=config, server=get_server('ACME-SEC')))
        await storage.put(IMAGE_URL % '5', IMAGE_BYTES)

        created = await storage.exists(IMAGE_URL % '5')
        self.assertTrue(created)

        await storage.remove(IMAGE_URL % '5')
        exists = await storage.exists(IMAGE_URL % '5')
        self.assertFalse(exists)

        await storage.put(IMAGE_URL % '5', IMAGE_BYTES)
        exists = await storage.exists(IMAGE_URL % '5')
        self.assertTrue(exists)

    def test_should_return_storage_prefix(self):
        config = Config(TC_AWS_STORAGE_BUCKET=s3_bucket, TC_AWS_STORAGE_ROOT_PATH='tata')
        storage = Storage(Context(config=config, server=get_server('ACME-SEC')))
        topic = storage._normalize_path('toto')
        self.assertEqual(topic, "tata/toto")

    def should_normalize_slash(self):
        config = Config(TC_AWS_STORAGE_ROOT_PATH='', TC_AWS_ROOT_IMAGE_NAME='root_image')
        storage = Storage(Context(config=config))
        self.assertEqual(storage._normalize_path('/test'), 'test')
        self.assertEqual(storage._normalize_path('/test/'), 'test/root_image')
        self.assertEqual(storage._normalize_path('/test/image.png'), 'test/image.png')


class CryptoS3StorageTestCase(S3MockedAsyncTestCase):

    @gen_test
    async def test_should_raise_on_invalid_config(self):
        config = Config(TC_AWS_STORAGE_BUCKET=s3_bucket, STORES_CRYPTO_KEY_FOR_EACH_IMAGE=True)
        storage = Storage(Context(config=config, server=get_server('')))

        await storage.put(IMAGE_URL % '9999', IMAGE_BYTES)

        with raises(RuntimeError, match='STORES_CRYPTO_KEY_FOR_EACH_IMAGE can\'t be True if no SECURITY_KEY specified'):
            await storage.put_crypto(IMAGE_URL % '9999')

    @gen_test
    async def test_getting_crypto_for_a_new_image_returns_none(self):
        config = Config(TC_AWS_STORAGE_BUCKET=s3_bucket, STORES_CRYPTO_KEY_FOR_EACH_IMAGE=True)
        storage = Storage(Context(config=config, server=get_server('ACME-SEC')))
        topic = await storage.get_crypto(IMAGE_URL % '9999')
        self.assertIsNone(topic)

    @gen_test
    async def test_does_not_store_if_config_says_not_to(self):
        config = Config(TC_AWS_STORAGE_BUCKET=s3_bucket, STORES_CRYPTO_KEY_FOR_EACH_IMAGE=False)
        storage = Storage(Context(config=config, server=get_server('ACME-SEC')))
        await storage.put(IMAGE_URL % '9998', IMAGE_BYTES)
        await storage.put_crypto(IMAGE_URL % '9998')
        topic = await storage.get_crypto(IMAGE_URL % '9998')
        self.assertIsNone(topic)

    @gen_test
    async def test_can_store_crypto(self):
        config = Config(TC_AWS_STORAGE_BUCKET=s3_bucket, STORES_CRYPTO_KEY_FOR_EACH_IMAGE=True)
        storage = Storage(Context(config=config, server=get_server('ACME-SEC')))
        await storage.put(IMAGE_URL % '6', IMAGE_BYTES)
        await storage.put_crypto(IMAGE_URL % '6')
        topic = await storage.get_crypto(IMAGE_URL % '6')

        self.assertIsNotNone(topic)
        self.assertNotIsInstance(topic, BaseException)
        self.assertEqual(topic, 'ACME-SEC')


class DetectorS3StorageTestCase(S3MockedAsyncTestCase):

    @gen_test
    async def test_can_store_detector_data(self):
        config = Config(TC_AWS_STORAGE_BUCKET=s3_bucket)
        storage = Storage(Context(config=config, server=get_server('ACME-SEC')))
        await storage.put(IMAGE_URL % '7', IMAGE_BYTES)
        await storage.put_detector_data(IMAGE_URL % '7', 'some-data')
        topic = await storage.get_detector_data(IMAGE_URL % '7')

        self.assertEqual(topic, 'some-data')

    @gen_test
    async def test_returns_none_if_no_detector_data(self):
        config = Config(TC_AWS_STORAGE_BUCKET=s3_bucket)
        storage = Storage(Context(config=config, server=get_server('ACME-SEC')))
        topic = await storage.get_detector_data(IMAGE_URL % '9999')

        self.assertIsNone(topic)


class WebpS3StorageTestCase(TestCase):

    def test_has_config_request(self):
        config = Config(AUTO_WEBP=True)
        context = Context(config=config)
        context.request = RequestParameters(accepts_webp=True)
        storage = Storage(context)
        self.assertTrue(storage.is_auto_webp)

    def test_has_config_no_request(self):
        config = Config(AUTO_WEBP=True)
        context = Context(config=config)
        storage = Storage(context)
        self.assertFalse(storage.is_auto_webp)

    def test_has_config_request_does_not_accept(self):
        config = Config(AUTO_WEBP=True)
        context = Context(config=config)
        context.request = RequestParameters(accepts_webp=False)
        storage = Storage(context)
        self.assertFalse(storage.is_auto_webp)

    def test_has_no_config(self):
        config = Config(AUTO_WEBP=False)
        context = Context(config=config)
        context.request = RequestParameters(accepts_webp=True)
        storage = Storage(context)
        self.assertFalse(storage.is_auto_webp)


class ExpiredTestCase(TestCase):

    @property
    def expired_enabled(self):
        return Storage(Context(config=Config(STORAGE_EXPIRATION_SECONDS=3600)))

    def test_should_check_invalid_key(self):
        self.assertTrue(self.expired_enabled.is_expired(None))
        self.assertTrue(self.expired_enabled.is_expired(False))
        self.assertTrue(self.expired_enabled.is_expired(dict()))
        self.assertTrue(self.expired_enabled.is_expired({'Error': ''}))

    def test_should_tell_when_not_expired(self):
        key = {
            'LastModified': datetime.now(tzutc()),
            'Body': 'foobar',
        }
        self.assertFalse(self.expired_enabled.is_expired(key))

    def test_should_tell_when_expired(self):
        key = {
            'LastModified': (datetime.now(tzutc()) - timedelta(seconds=3601)),
            'Body': 'foobar',
        }
        self.assertTrue(self.expired_enabled.is_expired(key))

    def test_expire_disabled_should_not_tell_when_expired(self):
        topic = Storage(Context(config=Config(STORAGE_EXPIRATION_SECONDS=0)))
        key = {
            'LastModified': (datetime.now(tzutc()) - timedelta(seconds=3601)),
            'Body': 'foobar',
        }
        self.assertFalse(topic.is_expired(key))



# coding: utf-8

# Copyright (c) 2015, thumbor-community
# Use of this source code is governed by the MIT license that can be
# found in the LICENSE file.

import botocore.session
from derpconf.config import Config
from mock import patch
from thumbor.context import Context
from thumbor.loaders import LoaderResult
from tornado.testing import gen_test

from .fixtures.storage_fixture import IMAGE_PATH, IMAGE_BYTES, s3_bucket
from tc_aws.loaders import s3_loader
from tests import S3MockedAsyncTestCase


class S3LoaderTestCase(S3MockedAsyncTestCase):

    @gen_test
    async def test_can_load_image(self):
        client = botocore.session.get_session().create_client('s3', endpoint_url='http://localhost:5000')

        client.put_object(
            Bucket=s3_bucket,
            Key=''.join(['root_path', IMAGE_PATH]),
            Body=IMAGE_BYTES,
            ContentType='image/jpeg', )

        conf = Config(
            TC_AWS_LOADER_BUCKET=s3_bucket,
            TC_AWS_LOADER_ROOT_PATH='root_path'
        )

        loader_result = await s3_loader.load(Context(config=conf), IMAGE_PATH)
        self.assertTrue(loader_result.successful)
        self.assertEqual(loader_result.buffer, IMAGE_BYTES)
        self.assertTrue('size' in loader_result.metadata)
        self.assertIsNone(loader_result.error)

    @gen_test
    async def test_returns_404_on_no_image(self):
        conf = Config(
            TC_AWS_LOADER_BUCKET=s3_bucket,
            TC_AWS_LOADER_ROOT_PATH='root_path'
        )

        loader_result = await s3_loader.load(Context(config=conf), 'foo-bar.jpg')
        self.assertFalse(loader_result.successful)
        self.assertIsNone(loader_result.buffer)
        self.assertEqual(loader_result.error, LoaderResult.ERROR_NOT_FOUND)

    @gen_test
    async def test_can_validate_buckets(self):
        conf = Config(
            TC_AWS_ALLOWED_BUCKETS=['whitelist_bucket'],
            TC_AWS_LOADER_BUCKET=None,
        )

        image = await s3_loader.load(Context(config=conf), '/'.join([s3_bucket, IMAGE_PATH]))
        self.assertIsNone(image.buffer)

    @patch('thumbor.loaders.http_loader.load')
    @gen_test
    async def test_should_use_http_loader(self, load_sync_patch):
        conf = Config(TC_AWS_ENABLE_HTTP_LOADER=True)
        await s3_loader.load(Context(config=conf), 'http://foo.bar')
        self.assertTrue(load_sync_patch.called)

    @patch('thumbor.loaders.http_loader.load')
    @gen_test
    async def test_should_not_use_http_loader_if_not_prefixed_with_scheme(self, load_sync_patch):
        conf = Config(TC_AWS_ENABLE_HTTP_LOADER=True)
        await s3_loader.load(Context(config=conf), 'foo/bar')
        self.assertFalse(load_sync_patch.called)



#coding: utf-8

# Copyright (c) 2015-2016, thumbor-community
# Use of this source code is governed by the MIT license that can be
# found in the LICENSE file.
from botocore.exceptions import BotoCoreError, ClientError
from thumbor.result_storages import BaseStorage, ResultStorageResult

from ..aws.storage import AwsStorage

from thumbor.utils import logger


class Storage(AwsStorage, BaseStorage):
    """
    S3 Result Storage
    """
    def __init__(self, context):
        """
        Constructor
        :param Context context: Thumbor's context
        """
        BaseStorage.__init__(self, context)
        AwsStorage.__init__(self, context, 'TC_AWS_RESULT_STORAGE')
        self.storage_expiration_seconds = context.config.get('RESULT_STORAGE_EXPIRATION_SECONDS', 3600)

    async def put(self, image_bytes):
        """
        Stores image
        :param bytes image_bytes: Data to store
        """
        path = self._normalize_path(self.context.request.url)

        metadata = {}

        if self.context.config.get('TC_AWS_STORE_METADATA'):
            metadata = dict(self.context.headers)

        try:
            return await self._put_object(image_bytes, path, metadata)
        except BotoCoreError as err:
            logger.exception('Unable to store result image object: %s', err)
            return None

    async def get(self, path = None):
        """
        Retrieves data
        :param string path: Path to load data (defaults to request URL)
        """
        if path is None:
            path = self.context.request.url


        try:
            key = await super(Storage, self).get(path)
        except ClientError:
            return None

        if key is None or self.is_expired(key):
            return None

        result = ResultStorageResult()
        async with key['Body'] as stream:
            result.buffer = await stream.read()
        result.successful = True

        result.metadata = {
            "LastModified": key.get('LastModified'),
            "ContentLength": key.get('ContentLength', None),
            "ContentType": key.get('ContentType'),
        }

        result.metadata.update(key.get('Metadata'))

        logger.debug(str(result.metadata))

        return result



# coding: utf-8

# Copyright (c) 2015, thumbor-community
# Use of this source code is governed by the MIT license that can be
# found in the LICENSE file.
from urllib.parse import unquote

import thumbor.loaders.http_loader as http_loader
from botocore.exceptions import ClientError
from thumbor.utils import logger
from thumbor.loaders import LoaderResult

from ..aws.bucket import Bucket


async def load(context, url):
    """
    Loads image
    :param Context context: Thumbor's context
    :param string url: Path to load
    """
    if _use_http_loader(context, url):
        return await http_loader.load(context, url)

    bucket, key = _get_bucket_and_key(context, url)

    if not _validate_bucket(context, bucket):
        result = LoaderResult(successful=False,
                              error=LoaderResult.ERROR_NOT_FOUND)
        return result

    loader = Bucket(
        bucket,
        context.config.get('TC_AWS_REGION'),
        context.config.get('TC_AWS_ENDPOINT'),
        context.config.get('TC_AWS_MAX_RETRY')
    )

    result = LoaderResult()

    try:
        file_key = await loader.get(key)
    except ClientError as err:
        logger.error(
            "ERROR retrieving image from S3 {0}: {1}".
                format(key, str(err.response)))

        # If we got here, there was a failure.
        # We will return 404 if S3 returned a 404, otherwise 502.
        result.successful = False

        if not err.response:
            result.error = LoaderResult.ERROR_UPSTREAM
            return result

        status_code = err.response.get('ResponseMetadata', {}).get('HTTPStatusCode')

        if status_code == 404:
            result.error = LoaderResult.ERROR_NOT_FOUND
            return result

        result.error = LoaderResult.ERROR_UPSTREAM
        return result

    result.successful = True
    async with file_key['Body'] as stream:
        result.buffer = await stream.read()

    result.metadata.update(
        size=file_key['ContentLength'],
        updated_at=file_key['LastModified'],
    )

    return result


def _get_bucket_and_key(context, url):
    """
    Returns bucket and key from url
    :param Context context: Thumbor's context
    :param string url: The URL to parse
    :return: A tuple with the bucket and the key detected
    :rtype: tuple
    """
    url = unquote(url)

    bucket = context.config.get('TC_AWS_LOADER_BUCKET')
    if not bucket:
        bucket = _get_bucket(url)
        url = '/'.join(url.lstrip('/').split('/')[1:])

    key = _get_key(url, context)

    return bucket, key


def _get_bucket(url):
    """
    Retrieves the bucket based on the URL
    :param string url: URL to parse
    :return: bucket name
    :rtype: string
    """
    url_by_piece = url.lstrip("/").split("/")

    return url_by_piece[0]


def _get_key(path, context):
    """
    Retrieves key from path
    :param string path: Path to analyze
    :param Context context: Thumbor's context
    :return: Extracted key
    :rtype: string
    """
    root_path = context.config.get('TC_AWS_LOADER_ROOT_PATH')
    return '/'.join([root_path, path]) if root_path is not '' else path


def _validate_bucket(context, bucket):
    """
    Checks that bucket is allowed
    :param Context context: Thumbor's context
    :param string bucket: Bucket name
    :return: Whether bucket is allowed or not
    :rtype: bool
    """
    allowed_buckets = context.config.get('TC_AWS_ALLOWED_BUCKETS', default=None)
    return not allowed_buckets or bucket in allowed_buckets


def _use_http_loader(context, url):
    """
    Should we use HTTP Loader with given path? Based on configuration as well.
    :param Context context: Thumbor's context
    :param string url: URL to analyze
    :return: Whether we should use HTTP Loader or not
    :rtype: bool
    """
    enable_http_loader = context.config.get('TC_AWS_ENABLE_HTTP_LOADER', default=False)
    return enable_http_loader and url.startswith('http')



# coding: utf-8

# Copyright (c) 2015, thumbor-community
# Use of this source code is governed by the MIT license that can be
# found in the LICENSE file.

from datetime import datetime, timedelta
from unittest import TestCase

from dateutil.tz import tzutc
from thumbor.config import Config
from thumbor.context import Context
from tornado.testing import gen_test

from .fixtures.storage_fixture import IMAGE_BYTES, get_server, s3_bucket
from tc_aws.result_storages.s3_storage import Storage
from tests import S3MockedAsyncTestCase


class Request(object):
    url = None


class S3StorageTestCase(S3MockedAsyncTestCase):

    @gen_test
    async def test_can_get_image(self):
        config = Config(TC_AWS_RESULT_STORAGE_BUCKET=s3_bucket)
        ctx = Context(config=config, server=get_server('ACME-SEC'))
        ctx.request = Request
        ctx.request.url = 'my-image-2.jpg'

        storage = Storage(ctx)
        await storage.put(IMAGE_BYTES)

        topic = await storage.get('my-image-2.jpg')

        self.assertEqual(topic.buffer, IMAGE_BYTES)

    @gen_test
    async def test_can_get_randomized_image(self):
        config = Config(TC_AWS_RESULT_STORAGE_BUCKET=s3_bucket, TC_AWS_RANDOMIZE_KEYS=True)
        ctx = Context(config=config, server=get_server('ACME-SEC'))
        ctx.request = Request
        ctx.request.url = 'my-image-2.jpg'

        storage = Storage(ctx)
        await storage.put(IMAGE_BYTES)

        topic = await storage.get()

        self.assertEqual(topic.buffer, IMAGE_BYTES)

    @gen_test
    async def test_can_get_image_with_metadata(self):
        config = Config(TC_AWS_RESULT_STORAGE_BUCKET=s3_bucket, TC_AWS_STORE_METADATA=True)
        ctx = Context(config=config, server=get_server('ACME-SEC'))
        ctx.headers = {'Some-Other-Header': 'doge-header'}
        ctx.request = Request
        ctx.request.url = 'my-image-meta.jpg'

        storage = Storage(ctx)
        await storage.put(IMAGE_BYTES)

        file_abspath = storage._normalize_path(ctx.request.url)
        topic = await storage.get(file_abspath)

        self.assertEqual(topic.metadata['some-other-header'], 'doge-header')
        self.assertEqual(topic.buffer, IMAGE_BYTES)


class ExpiredTestCase(TestCase):

    @property
    def expired_enabled(self):
        return Storage(Context(config=Config(RESULT_STORAGE_EXPIRATION_SECONDS=3600)))

    def test_should_check_invalid_key(self):
        self.assertTrue(self.expired_enabled.is_expired(None))
        self.assertTrue(self.expired_enabled.is_expired(False))
        self.assertTrue(self.expired_enabled.is_expired(dict()))
        self.assertTrue(self.expired_enabled.is_expired({'Error': ''}))

    def test_should_tell_when_not_expired(self):
        key = {
            'LastModified': datetime.now(tzutc()),
            'Body': 'foobar',
        }
        self.assertFalse(self.expired_enabled.is_expired(key))

    def test_should_tell_when_expired(self):
        key = {
            'LastModified': (datetime.now(tzutc()) - timedelta(seconds=3601)),
            'Body': 'foobar',
        }
        self.assertTrue(self.expired_enabled.is_expired(key))

    def test_expire_disabled_should_not_tell_when_expired(self):
        topic = Storage(Context(config=Config(RESULT_STORAGE_EXPIRATION_SECONDS=0)))
        key = {
            'LastModified': (datetime.now(tzutc()) - timedelta(seconds=3601)),
            'Body': 'foobar',
        }
        self.assertFalse(topic.is_expired(key))


class PrefixTestCase(TestCase):

    def test_should_check_invalid_key(self):
        config = Config(TC_AWS_RESULT_STORAGE_BUCKET=s3_bucket, TC_AWS_RESULT_STORAGE_ROOT_PATH='tata')
        ctx = Context(config=config, server=get_server('ACME-SEC'))

        storage = Storage(ctx)

        topic = storage._normalize_path('toto')

        self.assertEqual(topic, "tata/toto")



# coding: utf-8

# Copyright (c) 2015, thumbor-community
# Use of this source code is governed by the MIT license that can be
# found in the LICENSE file.

import aiobotocore
from botocore.client import Config
from thumbor.utils import logger
from thumbor.engines import BaseEngine


class Bucket(object):
    _client = None
    _instances = {}

    @staticmethod
    def __new__(cls, bucket, region, endpoint, *args, **kwargs):
        key = (bucket, region, endpoint) + args + tuple(kwargs.items())

        if not cls._instances.get(key):
            cls._instances[key] = super(Bucket, cls).__new__(cls)

        return cls._instances[key]

    """
    This handles all communication with AWS API
    """
    def __init__(self, bucket, region, endpoint, max_retry=None):
        """
        Constructor
        :param string bucket: The bucket name
        :param string region: The AWS API region to use
        :param string endpoint: A specific endpoint to use
        :return: The created bucket
        """
        self._bucket = bucket

        config = None
        if max_retry is not None:
            config = Config(
                retries=dict(
                    max_attempts=max_retry
                )
            )

        if self._client is None:
            self._client = aiobotocore.get_session().create_client(
                's3',
                region_name=region,
                endpoint_url=endpoint,
                config=config
            )

    async def exists(self, path):
        """
        Checks if an object exists at a given path
        :param string path: Path or 'key' to retrieve AWS object
        """
        try:
            await self._client.head_object(
                Bucket=self._bucket,
                Key=self._clean_key(path),
            )
        except Exception:
            return False
        return True

    async def get(self, path):
        """
        Returns object at given path
        :param string path: Path or 'key' to retrieve AWS object
        """

        return await self._client.get_object(
            Bucket=self._bucket,
            Key=self._clean_key(path),
        )

    async def get_url(self, path, method='GET', expiry=3600):
        """
        Generates the presigned url for given key & methods
        :param string path: Path or 'key' for requested object
        :param string method: Method for requested URL
        :param int expiry: URL validity time
        """

        url = await self._client.generate_presigned_url(
            ClientMethod='get_object',
            Params={
                'Bucket': self._bucket,
                'Key': self._clean_key(path),
            },
            ExpiresIn=expiry,
            HttpMethod=method,
        )

        return url

    async def put(self, path, data, metadata=None, reduced_redundancy=False, encrypt_key=False):
        """
        Stores data at given path
        :param string path: Path or 'key' for created/updated object
        :param bytes data: Data to write
        :param dict metadata: Metadata to store with this data
        :param bool reduced_redundancy: Whether to reduce storage redundancy or not?
        :param bool encrypt_key: Encrypt data?
        """
        storage_class = 'REDUCED_REDUNDANCY' if reduced_redundancy else 'STANDARD'
        content_type = BaseEngine.get_mimetype(data) or 'application/octet-stream'

        args = dict(
            Bucket=self._bucket,
            Key=self._clean_key(path),
            Body=data,
            ContentType=content_type,
            StorageClass=storage_class,
        )

        if encrypt_key:
            args['ServerSideEncryption'] = 'AES256'

        if metadata is not None:
            args['Metadata'] = metadata

        return await self._client.put_object(**args)

    async def delete(self, path):
        """
        Deletes key at given path
        :param string path: Path or 'key' to delete
        """
        return await self._client.delete_object(
            Bucket=self._bucket,
            Key=self._clean_key(path),
        )

    def _clean_key(self, path):
        logger.debug('Cleaning key: {path!r}'.format(path=path))
        key = path
        while '//' in key:
            logger.debug(key)
            key = key.replace('//', '/')

        if '/' == key[0]:
            key = key[1:]

        logger.debug('Cleansed key: {key!r}'.format(key=key))
        return key



# coding: utf-8

# Copyright (c) 2015-2016, thumbor-community
# Use of this source code is governed by the MIT license that can be
# found in the LICENSE file.

from os.path import join
from datetime import datetime
from dateutil.tz import tzutc
from hashlib import sha1

from .bucket import Bucket

class AwsStorage():
    """
    Base storage class
    """
    @property
    def is_auto_webp(self):
        """
        Determines based on context whether we automatically use webp or not
        :return: Use WebP?
        :rtype: bool
        """
        return self.context.config.AUTO_WEBP and hasattr(self.context, 'request') and self.context.request.accepts_webp

    @property
    def storage(self):
        """
        Instantiates bucket based on configuration
        :return: The bucket
        :rtype: Bucket
        """
        return Bucket(self._get_config('BUCKET'), self.context.config.get('TC_AWS_REGION'),
                      self.context.config.get('TC_AWS_ENDPOINT'))

    def __init__(self, context, config_prefix):
        """
        Constructor
        :param Context context: An instance of thumbor's context
        :param string config_prefix: Prefix used to load configuration values
        """
        self.config_prefix = config_prefix
        self.context = context

    async def get(self, path):
        """
        Gets data at path
        :param string path: Path for data
        """
        file_abspath = self._normalize_path(path)

        return await self.storage.get(file_abspath)

    def is_expired(self, key):
        """
        Tells whether key has expired
        :param  key: Path to check
        :return: Whether it is expired or not
        :rtype: bool
        """
        if key and 'LastModified' in key:
            expire_in_seconds = self.storage_expiration_seconds

            # Never expire
            if expire_in_seconds is None or expire_in_seconds == 0:
                return False

            timediff = datetime.now(tzutc()) - key['LastModified']

            return timediff.seconds > expire_in_seconds
        else:
            #If our key is bad just say we're expired
            return True

    async def _put_object(self, object_data, path, metadata=None):
        """
        Stores data at given path
        :param bytes bytes: Data to store
        :param string abspath: Path to store the data at
        :return: Path where the data is stored
        :rtype: string
        """

        return await self.storage.put(
            path,
            object_data,
            metadata=metadata,
            reduced_redundancy=self.context.config.get('TC_AWS_STORAGE_RRS', False),
            encrypt_key=self.context.config.get('TC_AWS_STORAGE_SSE', False),
        )

    def _get_config(self, config_key):
        """
        Retrieve specific config based on prefix
        :param string config_key: Requested config
        :param default: Default value if not found
        :return: Resolved config value
        """
        return getattr(self.context.config, '%s_%s' % (self.config_prefix, config_key))

    def _normalize_path(self, path):
        """
        Adapts path based on configuration (root_path for instance)
        :param string path: Path to adapt
        :return: Adapted path
        :rtype: string
        """
        path = path.lstrip('/')  # Remove leading '/'
        path_segments = [path]

        root_path = self._get_config('ROOT_PATH')
        if root_path and root_path is not '':
            path_segments.insert(0, root_path)

        if self.is_auto_webp:
            path_segments.append("webp")

        if self._should_randomize_key():
            path_segments.insert(0, self._generate_digest(path_segments))

        normalized_path = join(path_segments[0], *path_segments[1:]).lstrip('/') if len(path_segments) > 1 else path_segments[0]
        if normalized_path.endswith('/'):
            normalized_path += self.context.config.TC_AWS_ROOT_IMAGE_NAME

        return normalized_path

    def _should_randomize_key(self):
        return self.context.config.TC_AWS_RANDOMIZE_KEYS

    def _generate_digest(self, segments):
        return sha1(".".join(segments).encode('utf-8')).hexdigest()
